""" MaxCharMatch metric. """

import datetime
import os
import string
from dataclasses import dataclass
from typing import Dict, List, Tuple

import datasets
import diff_match_patch as dmp_module
import numpy as np
from more_itertools import chunked
from tqdm import tqdm

from max_char_match.alignment import SpellAlignment

_CITATION = """\
@misc{havens2019mcm,
    title  = {MaxCharMatch metric},
    author = {Sam Havens, Aneta Melisa Stal},
    year   = {2021}
}
"""

_DESCRIPTION = """\
MaxCharMatch metric is adaptation of MaxMatch (M^2) metrics (https://github.com/nusnlp/m2scorer) for spelling.
"""

_KWARGS_DESCRIPTION = """
Compute MaxCharMatch.
Args:
    predictions: list of predictions to score.
    references: list of references.
Returns:
    "tp": True Positive Count
    "fp": False Positive Count
    "fn": False Negative Count
    "precision": Precision
    "recall": Recall
    "F05": F0.5
Examples:
    >>> mcm_metric = datasets.load_metric("mcm.py")
    >>> mcm_metric._load_data(["Thiz si cool"], ["This is cool"])
    >>> mcm_metric._save_to(path/to/some/dir)
    >>> mcm_metric.compute(predictions=["This si cools"], references=["This is cool"])
    {'tp': 2, 'fp': 2, 'fn': 2, 'precision': 0.5, 'recall': 0.5, 'f05': 0.5}
"""


@dataclass
class EvalExample:
    sentence: str
    norm_form: str = ""

    def __post_init__(self):
        self.norm_form = self.normalize_sentence(sentence)

    @staticmethod
    def normalize_sentence(sent: str) -> str:
        sent = sent.strip().lower()
        return sent.translate(str.maketrans("", "", string.punctuation))

    @staticmethod
    def label_sentence(sent: str) -> str:
        return sent


@datasets.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MaxCharMatch(datasets.Metric):
    def __init__(self, *args, **kwargs):
        super(MaxCharMatch, self).__init__(*args, **kwargs)
        self.sa = SpellAlignment()
        self.save = False
        self.tokenizer = None
        self.batch_size = 16
        self.origs = []
        self.cors = []

    @staticmethod
    def normalize_sentence(sent: str) -> str:
        sent = sent.strip().lower()
        return sent.translate(str.maketrans("", "", string.punctuation))

    def convert_to_text(self, inputs) -> List[str]:
        text = []
        num_batches = int(len(inputs) / self.batch_size)
        for chunk in tqdm(chunked(inputs, self.batch_size), total=num_batches):
            text += self.tokenizer.batch_decode(
                chunk,
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
        return text

    def generator_to_text(self, inputs_generator) -> Tuple[List[str], List[str]]:
        inputs = []
        references = []
        for i in tqdm(range(len(inputs_generator))):
            sample = inputs_generator[i]
            inputs.append(sample["input_ids"])
            references.append(sample["labels"])
        inputs_text = self.convert_to_text(inputs)
        references_text = self.convert_to_text(references)
        return inputs_text, references_text

    def _load_data(self, data_generator) -> None:
        self.tokenizer = data_generator.tokenizer
        self.origs, self.cors = self.generator_to_text(data_generator)

    def _save_to(self, output_dir: str) -> None:
        self.save = True
        self.dmp = dmp_module.diff_match_patch()
        self.output_dir = os.path.join(output_dir, "eval_outputs")
        os.makedirs(self.output_dir, exist_ok=True)

    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("string"),
                    "references": datasets.Value("string"),
                }
            ),
            codebase_urls=[],
            reference_urls=[],
            format="numpy",
        )

    def compare_edits(self, orig: str, cor: str, pred: str) -> Tuple[int, int, int]:
        orig = self.normalize_sentence(orig)
        cor = self.normalize_sentence(cor)
        pred = self.normalize_sentence(pred)

        pred_edits = self.sa(orig, pred)
        ref_edits = self.sa(orig, cor)

        tp = len([pred_edit for pred_edit in pred_edits if pred_edit in ref_edits])
        fp = len(pred_edits) - tp
        fn = len([ref_edit for ref_edit in ref_edits if ref_edit not in pred_edits])

        return tp, fp, fn

    def get_html_diff(self, sent1: str, sent2: str) -> str:
        diff = self.dmp.diff_main(sent1, sent2)
        self.dmp.diff_cleanupSemantic(diff)
        return self.dmp.diff_prettyHtml(diff)

    def _compute(
        self, predictions: List[str], references: List[str]
    ) -> Dict[str, float]:
        tp_agg, fp_agg, fn_agg = 0, 0, 0

        f_out = None

        if self.save:
            f_name = os.path.join(
                self.output_dir,
                datetime.datetime.now().strftime("%d-%m-%Y-%H:%M:%S") + ".html",
            )
            f_out = open(f_name, "w", encoding="utf8")

        for o, c, p in tqdm(
            zip(self.origs, self.cors, predictions), total=len(self.origs)
        ):
            tp, fp, fn = self.compare_edits(o, c, p)
            tp_agg += tp
            fp_agg += fp
            fn_agg += fn

            if f_out:
                f_out.write(self.get_html_diff(o, c) + "<br>\n")
                f_out.write(self.get_html_diff(o, p) + "<br><br>\n\n")

        f_out.close()

        precision = tp_agg / (fp_agg + tp_agg) if fp_agg * tp_agg else 0
        recall = tp_agg / (tp_agg + fn_agg) if tp_agg * fn_agg else 0
        f05 = (
            1.25 * precision * recall / (0.25 * precision + recall)
            if precision * recall
            else 0
        )
        f2 = (
            (5 * precision * recall) / (4 * precision + recall)
            if precision * recall
            else 0
        )

        return {
            "tp": tp_agg,
            "fp": fp_agg,
            "fn": fn_agg,
            "precision": precision,
            "recall": recall,
            "f05": f05,
            "f2": f2,
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['max_char_match']

package_data = \
{'': ['*']}

install_requires = \
['datasets>=2.1.0,<3.0.0',
 'diff-match-patch>=20200713,<20200714',
 'more-itertools>=8.12.0,<9.0.0',
 'numpy>=1.22.3,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

setup_kwargs = {
    'name': 'max-char-match',
    'version': '0.1.2',
    'description': 'MaxCharMatch metric',
    'long_description': None,
    'author': 'melisa-writer',
    'author_email': 'melisa@writer.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
This code comes partially from https://github.com/chrisjbryant/errant/blob/master/errant/alignment.py
"""
import re
from dataclasses import dataclass, field
from itertools import groupby


@dataclass
class CharSpan:
    start: int
    end: int
    sent: str = field(repr=False, compare=False)
    text: str = field(default="")

    def __post_init__(self):
        self.text = self.sent[self.start : self.end]


class CharDiff:
    def __init__(self, orig, cor, op):
        self.err_span = CharSpan(op[0], op[1], orig)
        self.cor_span = CharSpan(op[2], op[3], cor)

    def __str__(self):
        return f"'{self.err_span.text}' -> '{self.cor_span.text}'"

    def __eq__(self, other):
        if self.err_span != other.err_span:
            return False
        if self.cor_span != other.cor_span:
            return False
        return True

    def is_del(self):
        return len(self.err_span.text) and not len(self.cor_span.text)


class SpellAlignment:

    qwerty = "1234567890-=qwertyuiop[]asdfghjkl;' zxcvbnm,./"
    qwerty_dict = {k: i for i, k in enumerate(list(qwerty)) if k != " "}

    def __init__(self):
        self.only_letters = re.compile(r"[^a-zA-Z\s]")

    def __call__(self, orig: str, cor: str):
        """
        Input:
        - orig: original sent
        - cor: corrected sent
        """
        self.orig = orig
        self.cor = cor

        # Align orig and cor and get the cost and op matrices
        self.cost_matrix, self.op_matrix = self.align()

        # Get the cheapest align sequence from the op matrix
        self.align_seq = self.get_cheapest_align_seq()

        return self.get_all_merge_edits()

    @staticmethod
    def is_left_border(char):
        return char in ["1", "q", "a", "z"]

    @staticmethod
    def is_right_border(char):
        return char in ["=", "]", "'", "?"]

    def is_keyboard_neighbor(self, a, b):

        if a not in self.qwerty:
            return False

        if b not in self.qwerty:
            return False

        if not self.is_left_border(a):
            if self.qwerty_dict[a] - self.qwerty_dict[b] == 1:
                return True
            if abs(self.qwerty_dict[a] - self.qwerty_dict[b]) == 11:
                return True

        if not self.is_right_border(a):
            if self.qwerty_dict[a] - self.qwerty_dict[b] == -1:
                return True
            if abs(self.qwerty_dict[a] - self.qwerty_dict[b]) == 12:
                return True

        if self.is_left_border(a):
            if self.qwerty_dict[a] - self.qwerty_dict[b] == 11:
                return True

        if self.is_right_border(a):
            if self.qwerty_dict[a] - self.qwerty_dict[b] == 12:
                return True

        return False

    def align(self):
        """
        Returns:
        - cost matrix and the
        - operation matrix of the alignment
        """
        o_len = len(self.orig)
        c_len = len(self.cor)

        o_low = list(self.orig.lower())
        c_low = list(self.cor.lower())

        cost_matrix = [[0.0 for j in range(c_len + 1)] for i in range(o_len + 1)]
        op_matrix = [["O" for j in range(c_len + 1)] for i in range(o_len + 1)]

        for i in range(1, o_len + 1):
            cost_matrix[i][0] = cost_matrix[i - 1][0] + 1
            op_matrix[i][0] = "D"
        for j in range(1, c_len + 1):
            cost_matrix[0][j] = cost_matrix[0][j - 1] + 1
            op_matrix[0][j] = "I"

        for i in range(o_len):
            for j in range(c_len):
                # Matches
                if self.orig[i] == self.cor[j]:
                    cost_matrix[i + 1][j + 1] = cost_matrix[i][j]
                    op_matrix[i + 1][j + 1] = "M"
                # Non-matches
                else:
                    del_cost = cost_matrix[i][j + 1] + 1
                    ins_cost = cost_matrix[i + 1][j] + 1
                    trans_cost = float("inf")
                    # Custom substitution
                    sub_cost = cost_matrix[i][j] + self.get_sub_cost(
                        self.orig[i], self.cor[j]
                    )
                    # Transpositions require >=2 tokens
                    # Traverse the diagonal while there is not a Match.
                    k = 1
                    while (
                        i - k >= 0
                        and j - k >= 0
                        and cost_matrix[i - k + 1][j - k + 1]
                        != cost_matrix[i - k][j - k]
                    ):
                        if sorted(o_low[i - k : i + 1]) == sorted(c_low[j - k : j + 1]):
                            trans_cost = cost_matrix[i - k][j - k] + k
                            break
                        k += 1

                    # Costs
                    costs = [trans_cost, sub_cost, ins_cost, del_cost]
                    # Get the index of the cheapest (first cheapest if tied)
                    l = costs.index(min(costs))
                    # Save the cost and the op in the matrices
                    cost_matrix[i + 1][j + 1] = costs[l]
                    if l == 0:
                        op_matrix[i + 1][j + 1] = "T" + str(k + 1)
                    elif l == 1:
                        op_matrix[i + 1][j + 1] = "S"
                    elif l == 2:
                        op_matrix[i + 1][j + 1] = "I"
                    else:
                        op_matrix[i + 1][j + 1] = "D"
        return cost_matrix, op_matrix

    def get_sub_cost(self, o, c):
        """
        o - input char
        c - output char
        Returns:
            char cost between 0 < x < 2
        """
        # Short circuit if the only difference is case
        if o.lower() == c.lower():
            return 0

        # If space to not space: dissimilar
        if (o == " " and c != o) or (c == " " and o != c):
            return 2
        # If alpha to non alpha: dissimilar
        if self.is_nonalpha(o) and not self.is_nonalpha(c):
            return 2
        if self.is_nonalpha(c) and not self.is_nonalpha(o):
            return 2

        # Fat finger
        if self.is_keyboard_neighbor(c, o):
            return 0.5

        return 1

    def is_nonalpha(self, flagged_text: str):
        non_letters = self.only_letters.findall(flagged_text)
        return non_letters != []

    def get_cheapest_align_seq(self):
        """
        Get the cheapest alignment sequence and indices from the op matrix
        D - deletions, I - insertions, S - substitutions, M - matching
        """
        i = len(self.op_matrix) - 1
        j = len(self.op_matrix[0]) - 1
        align_seq = []
        while i + j != 0:

            op = self.op_matrix[i][j]
            if op in {"M", "S"}:
                align_seq.append((op, i - 1, i, j - 1, j))
                i -= 1
                j -= 1
            elif op == "D":
                align_seq.append((op, i - 1, i, j, j))
                i -= 1
            elif op == "I":
                align_seq.append((op, i, i, j - 1, j))
                j -= 1
            else:
                k = int(op[1:])
                align_seq.append((op, i - k, i, j - k, j))
                i -= k
                j -= k
        align_seq.reverse()
        return align_seq

    def edit_to_type(self, edit):
        """
        P (pass) - M or punct
        W (word correction) - not P and not space
        """
        is_matched = edit[0] == "M"

        edit_type = "W"

        if is_matched:
            edit_type = "P"

        return edit_type

    def get_all_merge_edits(self):
        """
        Merge all adjacent non-match ops
        accept punct-punct corrections but leave out other-punct
        spelling can't add, remove or replace by something else purely punct tokens
        """
        edits = []
        for op, group in groupby(self.align_seq, lambda x: self.edit_to_type(x)):
            if op != "P":
                merged = self.merge_edits(list(group))
                edits.append(CharDiff(self.orig, self.cor, merged[0][1:]))
        return edits

    def merge_edits(self, seq):
        """
        Merge the input alignment sequence to a single edit span
        """
        if seq:
            return [("X", seq[0][1], seq[-1][2], seq[0][3], seq[-1][4])]
        else:
            return seq


if __name__ == "__main__":

    sa = SpellAlignment()
    orig = "thiz is hutright, ro no!"
    cor = "this is hot right, or not?"
    edits = sa(orig, cor)
    for e in edits:
        print(e)
